package ch.zhaw.projekt2.turnierverwaltung;

import javafx.collections.FXCollections;
import javafx.collections.ObservableList;

import java.io.File;
import java.io.Serializable;
import java.util.*;

public class Tournament implements Serializable {
    private String name;
    private Type type;
    private List<Participant> participants;
    private List<Place> places;
    private List<List<Game>> gameList;


    public Tournament(String name, Type type) throws InvalidNameException, InvalidTypeException {
        if (!name.matches("[\\w]{1,20}")) {
            throw new InvalidNameException("Invalid Name entered"); //TODO handle en logging.
        } else if (!Arrays.asList(Type.values()).contains(type)) {
            throw new InvalidTypeException("Invalid Type selected"); //TODO handle en logging.
        }

        setName(name);
        setType(type);
        participants = new ArrayList<>();
        places = new ArrayList<>();
        gameList = new ArrayList<>();
    }

    public void addParticipant(Participant newParticipant) throws ParticipantExistsException {
        participants.removeIf(participant -> participant.equals(newParticipant));
        participants.add(newParticipant);
    }

    public void removeParticipant(Participant participant) throws ParticipantNotExistsException {
        if (!participants.contains(participant)) {
            throw new ParticipantNotExistsException("The given Participant is not part of this Tournament");
        }
        participants.remove(participant);
    }

    public ObservableList<Participant> getParticipants() {
        ObservableList<Participant> participantsObservable = FXCollections.observableArrayList();
        participantsObservable.addAll(participants);
        return participantsObservable;
    }

    public void addPlace(Place newPlace) throws PlaceExistsException {
        places.removeIf(place -> place.equals(newPlace));
        places.add(newPlace);
    }

    public void removePlace(Place place) throws PlaceNotExistsException {
        if (!places.contains(place)) {
            throw new PlaceNotExistsException("The given Place is not part of this Tournament");
        }
        places.remove(place);
    }

    public ObservableList<Place> getPlaces() {
        ObservableList<Place> placesObservable = FXCollections.observableArrayList();
        placesObservable.addAll(places);
        return placesObservable;
    }


    /*
    creates a complete new GameSchedule
     */
    public void createGameSchedule() throws NumberOfParticipantInvalidException {
        gameList = new ArrayList<>();
        if (type == Type.KO) {
            if (numberOfParticipantValid()) {
                calcGameSchedule();
                //TODO Logging
            } else {
                throw new NumberOfParticipantInvalidException("Can not Create Game Schedule for KO Modus");
                //TODO Logging
            }


        } else {
            //TODO for Type Group
        }
    }

    private boolean numberOfParticipantValid() {
        double res = Math.log(participants.size()) / Math.log(2);
        return (res * 10) % 10 == 0 && participants.size() >=4;
    }

    private void calcGameSchedule() {
        Collections.shuffle(participants);

        for (int i = 0; i < (Math.log(participants.size()) / Math.log(2)); i++) {
            List<Game> gameRound = new ArrayList<>();
            if (i == 0) {
                for (int j = 0; j < participants.size() - 1; j += 2) {
                    gameRound.add(new Game(participants.get(j), participants.get(j+1)));
                }
            } else {
                for (int j = 0; j < (gameList.get(i-1).size()/2); j++) {
                    gameRound.add(new Game(null,null));
                }
            }
            gameList.add(gameRound);
        }

    }

    public void refreshGameParticipants(){
        for(int i = 1; i < gameList.size(); i++){
            for(int j = 0; j/2 < gameList.get(i).size(); j ++){
                Participant winner = gameList.get(i-1).get(j).getWinner();
                if(winner != null){
                    if(j % 2 == 0){
                        gameList.get(i).get(j/2).setParticipant1(winner);
                    } else{
                        gameList.get(i).get(j/2).setParticipant2(winner);
                    }
                }
            }
        }
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public Type getType() {
        return type;
    }

    public void setType(Type type) {
        this.type = type;
    }

    public List<List<Game>> getGameList() {
        return gameList;
    }

    public enum Type {
        KO("KO-System"), GROUPS("Gruppenspiele");

        private String name;

        private Type(String name) {
            this.name = name;
        }

        @Override
        public String toString() {
            return name;
        }

        public static ObservableList<Type> getObservableList() {
            ObservableList<Type> items = FXCollections.observableArrayList();
            items.addAll(values());
            return items;
        }
    }

    public class InvalidTypeException extends Exception {
        public InvalidTypeException() {
            super();
        }

        public InvalidTypeException(String errorMessage) {
            super(errorMessage);
        }

    }

    public class ParticipantExistsException extends Exception {
        public ParticipantExistsException() {
            super();
        }

        public ParticipantExistsException(String errorMessage) {
            super(errorMessage);
        }

    }

    public class ParticipantNotExistsException extends Exception {
        public ParticipantNotExistsException() {
            super();
        }

        public ParticipantNotExistsException(String errorMessage) {
            super(errorMessage);
        }

    }

    public class PlaceExistsException extends Exception {
        public PlaceExistsException() {
            super();
        }

        public PlaceExistsException(String errorMessage) {
            super(errorMessage);
        }

    }

    public class PlaceNotExistsException extends Exception {
        public PlaceNotExistsException() {
            super();
        }

        public PlaceNotExistsException(String errorMessage) {
            super(errorMessage);
        }

    }

    public class NumberOfParticipantInvalidException extends Exception {
        public NumberOfParticipantInvalidException() {
            super();
        }

        public NumberOfParticipantInvalidException(String msg) {
            super(msg);
        }
    }

}
